#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct tableitem {
    char *option; /* option expected */
    int flag ;    /* which packages are needed */
    char *include;/* corresponding include file */
    char *main;   /* corresponding main function */
    char *init;   /* corresponding init function */
} tableitem;

static tableitem table[] = {
{"-tcl" , 1, "tcl.h"  ,"Tcl_Main"   ,"Tcl_InitStandAlone"},
{"-itcl", 3, "itcl.h" ,(char *)NULL ,"Itcl_Init"},
{"-pvm" , 5, "tkPvm.h","Pvm_Main"   ,"Pvm_InitStandAlone"},
{"-tk"  , 9, "tk.h"   ,"Tk_Main"    ,"Tk_InitStandAlone"},
{"-tix" ,25, "tix.h"  ,"Tix_Main"   ,"Tix_Init"},
};

static char *verbose = "\
*************************** tcl2c ********************************\n\
written by: Jan Nijtmans\n\
            NICI (Nijmegen Institute of Cognition and Information)\n\
	    email: nijtmans@nici.kun.nl\n\
	    url:   http://www.nici.kun.nl/~nijtmans/\n\n\
usage:	tcl2c -o file source1 source2 ?options?\n\
	tcl2c -help\n\
";

static char *help = "\
available options:\n\
	-o filename	output file (default is stdout)\n\
	-tcl		use Tcl\n\
	-itcl		use Itcl\n\
	-pvm		use tkPvm\n\
	-tk		use Tk\n\
	-tix		use Tix\n\n\
Other command line arguments are assumed to be tcl script-files. It is \n\
possible to include c-files (with the extension .c), which are already\n\
converted tcl-scripts. These are included using the \"#include\".\n\n\
The output file can be compiled with any c or c++ compiler.\n\
";

static char *part1 = "\
#include \"tcl.h\"\n\
";

static char *part2 = "\
EXTERN void		%s _ANSI_ARGS_((int argc, char **argv,\n\
			    Tcl_AppInitProc *appInitProc));\n\
static char *script[] = {\n\
";

static char *part3 = "\
(char *) NULL\n\
};\n\
EXTERN int matherr();\n\
int (*tclDummyMathPtr)() = matherr;\n\
#ifdef __cplusplus\n\
main (int    argc,		/* Number of command-line arguments. */\n\
      char **argv)		/* Values of command-line arguments. */\n\
#else\n\
main(argc, argv)\n\
    int argc;			/* Number of command-line arguments. */\n\
    char **argv;		/* Values of command-line arguments. */\n\
#endif\n\
{\n\
    %s(argc, argv, Tcl_AppInit);\n\
    return 0;			/* Needed only to prevent compiler warning. */\n\
}\n\
int\n\
#ifdef __cplusplus\n\
Tcl_AppInit (Tcl_Interp *interp)/* Interpreter for application. */\n\
#else\n\
Tcl_AppInit(interp)\n\
    Tcl_Interp *interp;		/* Interpreter for application. */\n\
#endif\n\
{\n\
    char **p = script;\n\
    Tcl_DString data;\n\
";

static char *part4 = "\
    Tcl_DStringInit(&data);\n\
    while(*p) {\n\
	Tcl_DStringSetLength(&data,0);\n\
	Tcl_DStringAppend(&data,*p++,-1);\n\
	if(Tcl_Eval(interp,Tcl_DStringValue(&data)) == TCL_ERROR) {\n\
	    Tcl_DStringFree(&data);\n\
	    return TCL_ERROR;\n\
	}\n\
    }\n\
    Tcl_DStringFree(&data);\n\
    tcl_RcFileName = \"~/.%src\";\n\
    return TCL_OK;\n\
}\n\
";

static char *initproto = "\
EXTERN int		%s _ANSI_ARGS_((Tcl_Interp *interp));\n\
";

static char *callinit = "\
    if (%s(interp) == TCL_ERROR) {\n\
        return TCL_ERROR;\n\
    }\n\
";

int printline(f,buffer)
    FILE *f;
    char *buffer;
{
    fputc('\"',f);
    while(*buffer) {
	if(*buffer=='\"'||*buffer=='\\') fputc('\\',f);
	if(*buffer=='\n') {fputc('\\',f); fputc('n',f); fputc('\\',f); }
	fputc(*buffer++,f);
    }
    fputc('\"',f);
    fputc(',',f);
    fputc('\n',f);
    return 0;
}

int main(argc, argv)
int argc;
char *argv[];
{
    FILE *fin,*fout;
    char *p,*q, *filename=NULL, *main, buffer[32768];
    tableitem *t;
    int c,i, bracelevel, flags=0;

    if(argc==1) {
	printf(verbose);
	exit(0);
    }
    if(argc==2&&!strcmp(argv[1],"-help")) {
	printf(verbose);
	printf(help);
	exit(0);
    }
/* parse all command line arguments */
    for(i=1; i<argc; i++) {
	if(!strcmp(argv[i],"-o")) {
	    i++; filename = argv[i];
	} else {
	    for(t=table;t<table+(sizeof(table)/sizeof(tableitem));t++) {
		if(!strcmp(argv[i],t->option)) {
		    flags |= t->flag;
		}
	    }
	}
    }
/* open output file, if not stdout */
    if(filename) {
	fout = fopen(filename,"w");
	if(fout==NULL) {
	    fprintf(stderr,"error opening file %s\n",filename);
	    exit(1);
	}
    } else {
	fout = stdout;
    }
/* create prototypes for all initialization functions that are used */
    if(flags) {
	fprintf(fout, part1);
	for(i=0,c=1;i<(sizeof(table)/sizeof(tableitem));i++,c<<=1) {
	    if(flags & c)
		fprintf(fout,initproto,table[i].init);
	}
	main="Tcl_Main";c=flags; i=0;
	while(c) {
	    if((c&1)&&(table[i].main)) main=table[i].main;
	    c>>=1; i++;
	}
	fprintf(fout, part2, main);
    }
/* handle all remaining arguments */
    if(argc) {argc--; argv++;}
    while(argc) {
	if((*argv)[0]=='-') {
	    if((*argv)[1]=='o'&&(*argv)[2]==0) {
		argc--; argv++;
	    }
	} else if((p=strstr(*argv,".c"))&&(p[2]==0)) {
	    fprintf(fout,"#include \"%s\"\n",*argv);
	} else {
	    fin=fopen(*argv,"r");
	    p = q = buffer; bracelevel=0;
	    while((c=fgetc(fin))!=EOF) {
		*p = c;
		if(p==buffer || p[-1]!='\\') {
		    switch(*p) {
			case '[':
			case '{':
			case '(': bracelevel+=2; break;
			case ']':
			case '}':
			case ')': bracelevel-=2; break;
			case '"': bracelevel^=1; break;
		    }
		}
		if(*p=='\n') {
		    if(p==buffer||*q=='\n'||*q=='#') {
			p=q;
		    } else if(bracelevel||p[-1]=='\\') {
			p++; q=p;
		    } else {
			*p = 0; printline(fout,buffer);
			p = q = buffer;
		    }
		} else {
		    p++;
		}
	    }
	    if(p!=buffer) {
		*p=0; printline(fout,buffer);
	    }
	    fclose(fin);
	}
	argc--; argv++;
    }
/* end of scripts, finally the functions main() and tclAppInit()  */
    if(flags) {
	fprintf(fout, part3,main);
	for(i=0,c=1;i<(sizeof(table)/sizeof(tableitem));i++,c<<=1) {
	    if(flags & c)
		fprintf(fout,callinit,table[i].init);
	}
	p=filename?filename:"app";
	if(q=strrchr(p,'/')) p=q+1;
	if(q=strchr(p,'.')) *q=0;
	if(!*p) p="app";
	fprintf(fout, part4,p);
    }
/* close output-file, if not stdout */
    if(fout!=stdout) {
	fclose(fout);
    }
    return 0;
}

